import logging
import os
from abc import ABC, abstractmethod

from jade.hpc.common import HpcType
from jade.jobs.pipeline_manager import PipelineManager
from jade.models import HpcConfig, LocalHpcConfig
from jade.models.pipeline import PipelineStage
from jade.utils.utils import dump_data, load_data
from disco.pipelines.enums import TemplateSection, TemplateParams
from disco.enums import AnalysisType

logger = logging.getLogger(__name__)


class PipelineTemplate:
    """class for handling pipeline template data"""
    
    def __init__(self, template_file):
        self.template_file = template_file
        self.data = load_data(template_file)
    
    @property
    def task_name(self):
        return self.data["task_name"]

    @property
    def inputs(self):
        return self.data["inputs"]

    @property
    def preconfigured(self):
        return self.data["preconfigured"]

    @property
    def simulation_type(self):
        return self.data["simulation_type"]
    
    @property
    def analysis_type(self):
        return self.data["analysis_type"]

    @property
    def database(self):
        return self.data["database"]

    @property
    def reports(self):
        return self.data.get(TemplateSection.REPORTS.value, {})

    def contains_prescreen(self):
        return TemplateSection.PRESCREEN.value in self.data

    def contains_postprocess(self):
        return TemplateSection.POSTPROCESS.value in self.data

    def update_transform_params(self, data):
        section = TemplateSection.MODEL
        _data = self._keep_null_value(data)
        self.data[section.value][TemplateParams.TRANSFORM_PARAMS.value].update(_data)

    def update_config_params(self, data, section):
        _data = self._keep_null_value(data)
        self.data[section.value][TemplateParams.CONFIG_PARAMS.value].update(_data)

    def update_reports_params(self, data):
        if TemplateSection.REPORTS.value not in self.data:
            self.data[TemplateSection.REPORTS.value] = {}
        _data = self._keep_null_value(data)
        self.data[TemplateSection.REPORTS.value].update(_data)
    
    @staticmethod
    def _keep_null_value(data):
        _data = {}
        for key, value in data.items():
            if value is None:
                _data[key] = "null"
            else:
                _data[key] = value
        return _data

    def get_command_params(self, section, params_type):
        """Return command params in dict"""
        if isinstance(section, TemplateSection):
            section = section.value
        if isinstance(params_type, TemplateParams):
            params_type = params_type.value
        
        if params_type not in self.data[section]:
            raise KeyError(f"Template section '{section}' does not contain '{params_type}'.")
        
        params = self.data[section].get(params_type)
        for param, value in params.items():
            if value == "null":
                params[param] = None
        return params

    def get_transform_params(self, section):
        return self.get_command_params(section, TemplateParams.TRANSFORM_PARAMS)

    def get_prescreen_params(self, section):
        return self.get_command_params(section, TemplateParams.PRESCREEN_PARAMS)

    def get_config_params(self, section):
        return self.get_command_params(section, TemplateParams.CONFIG_PARAMS)

    def get_submitter_params(self, section):
        return self.get_command_params(section, TemplateParams.SUMITTER_PARAMS)

    def get_command_options(self, section, params_type):
        """Return command options in constructed string"""
        if isinstance(section, TemplateSection):
            section = section.value
        if isinstance(params_type, TemplateParams):
            params_type = params_type.value
        
        params = self.data[section][params_type]
        options = self._construct_options_string(params)
        return options
    
    @staticmethod
    def _construct_options_string(params):
        temp = []
        for param, value in params.items():
            if value == "null" or value is False or value is None:
                continue
            dash_param = "--" + param.replace("_", "-")
            if value is True:
                temp.append(dash_param)
            else:
                temp.append(f"{dash_param}={value}")
        options =  " ".join(temp)
        return options
    
    def get_model_transform_output(self):
        if "JADE_PIPELINE_OUTPUT_DIR" in os.environ:
            pipeline_output = os.environ["JADE_PIPELINE_OUTPUT_DIR"]
        else:
            pipeline_output = "$JADE_PIPELINE_OUTPUT_DIR"
        params = self.get_transform_params(TemplateSection.MODEL)
        output = os.path.join(pipeline_output, params["output"])
        return output
    
    def get_transform_options(self, section):
        params = self.get_transform_params(TemplateSection.MODEL)
        params["output"] = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", params["output"])
        options = self._construct_options_string(params)
        return options
    
    def get_config_options(self, section):
        return self.get_command_options(section, TemplateParams.CONFIG_PARAMS)

    def get_submitter_options(self, section):
        return self.get_command_options(section, TemplateParams.SUMITTER_PARAMS)

    def remove_section(self, section):
        if isinstance(section, TemplateSection):
            section = section.value
        
        if section in self.data:
            self.data.pop(section)

    def remove_params(self, section, params_type):
        if isinstance(section, TemplateSection):
            section = section.value
        if isinstance(params_type, TemplateParams):
            params_type = params_type.value
        if section in self.data and params_type in self.data[section]:
            self.data[section].pop(params_type)

    def set_preconfigured_models(self, path):
        self.data[TemplateSection.MODEL.value]["preconfigured_models"] = path


class PipelineCreatorBase(ABC):
    """A base class for pipeline creator"""

    def __init__(self, template_file):
        self.template_file = template_file
        self.stage_num = 0
    
    @property
    def template(self):
        return PipelineTemplate(self.template_file)
    
    @staticmethod
    def get_auto_config_python_file():
        return os.path.join(os.path.dirname(__file__), "auto_config.py")
    
    def get_prescreen_auto_config_text_file(self):
        return os.path.join(".", "pipeline-prescreen-auto-config.txt")
    
    def get_simulation_auto_config_text_file(self):
        return os.path.join(".", "pipeline-simulation-auto-config.txt")

    def get_postprocess_auto_config_text_file(self):
        return os.path.join(".", "pipeline-postprocess-auto-config.txt")
    
    def get_postprocess_command_text_file(self):
        return os.path.join(".", "pipeline-postprocess-command.txt")
    
    @abstractmethod
    def create_pipeline(self, pipeline_config_file):
        """Create pipeline config file"""
    
    @abstractmethod
    def make_model_transform_command(self):
        """Make disco transform-model command"""
    
    @abstractmethod
    def make_prescreen_create_command(self):
        """Make disco prescreen-pv-penetration-levels command"""
    
    @abstractmethod
    def make_prescreen_filter_command(self):
        """Make disco prescreen-pv-penetration-levels filter-config command"""
    
    @abstractmethod
    def make_disco_config_command(self, section):
        """Make disco config command"""
    
    @abstractmethod
    def make_postprocess_command(self):
        """Make disco make-summary-tables & compute-hosting-capacity command"""
    
    def create_prescreen_auto_config_text_file(self):
        """Create script for generating prescreen config file"""
        temp = []
        if not self.template.preconfigured:
            transform_command = self.make_model_transform_command()
            temp.append(transform_command)
        config_command = self.make_disco_config_command(TemplateSection.PRESCREEN)
        prescreen_command = self.make_prescreen_create_command()
        temp.extend([config_command, prescreen_command])
        auto_config_command = "\n".join(temp)
        
        text_file = self.get_prescreen_auto_config_text_file()
        with open(text_file, "w") as f:
            f.write(auto_config_command)
            f.write("\n")
        return text_file
    
    def create_simulation_auto_config_text_file(self):
        """Create script for generating disco config file"""
        if self.template.contains_prescreen():
            auto_config_command = self.make_prescreen_filter_command()
        else:
            temp = []
            if not self.template.preconfigured:
                temp.append(self.make_model_transform_command())
            temp.append(self.make_disco_config_command(TemplateSection.SIMULATION))
            auto_config_command = "\n".join(temp)
        
        text_file = self.get_simulation_auto_config_text_file()
        with open(text_file, "w") as f:
            f.write(auto_config_command)
            f.write("\n")
        return text_file

    def create_postprocess_auto_config_text_file(self):
        """Create script for generating postprocess config file"""
        command_file = self.create_postprocess_command_text_file()
        options = self.template.get_config_options(TemplateSection.POSTPROCESS)
        auto_config_command = f"jade config create {command_file} {options}"
        
        text_file = self.get_postprocess_auto_config_text_file()
        with open(text_file, "w") as f:
            f.write(auto_config_command)
            f.write("\n")
        return text_file
    
    def create_postprocess_command_text_file(self):
        text_file = "pipeline-postprocess-command.txt"
        command = self.make_postprocess_command()
        with open(text_file, "w") as f:
            f.write(command)
            f.write("\n")
        return text_file

    def make_prescreen_stage(self):
        self.stage_num += 1
        auto_config_text_file = self.create_prescreen_auto_config_text_file()
        submitter_params = self.template.get_submitter_params(TemplateSection.PRESCREEN)
        submitter_params["hpc_config"] = _create_hpc_config(submitter_params)
        
        auto_config_py = self.get_auto_config_python_file()
        auto_config_command = f"python {auto_config_py} {auto_config_text_file}"
        prescreen_params = self.template.get_prescreen_params(TemplateSection.PRESCREEN)
        stage = PipelineStage(
            auto_config_cmd=auto_config_command,
            config_file=prescreen_params["prescreen_config_file"],
            stage_num=self.stage_num,
            submitter_params=submitter_params,
        )
        return stage

    def make_simulation_stage(self):
        self.stage_num += 1
        auto_config_text_file = self.create_simulation_auto_config_text_file()
        submitter_params = self.template.get_submitter_params(TemplateSection.SIMULATION)
        submitter_params["hpc_config"] = _create_hpc_config(submitter_params)
        
        if self.template.contains_prescreen():
            prescreen_params = self.template.get_prescreen_params(TemplateSection.PRESCREEN)
            config_file = prescreen_params["filtered_config_file"]
        else:
            config_params = self.template.get_config_params(TemplateSection.SIMULATION)
            config_file = config_params["config_file"]

        auto_config_py = self.get_auto_config_python_file()
        auto_config_command = f"python {auto_config_py} {auto_config_text_file}"
        stage = PipelineStage(
            auto_config_cmd=auto_config_command,
            config_file=config_file,
            stage_num=self.stage_num,
            submitter_params=submitter_params
        )
        return stage
    
    def make_postprocess_stage(self):
        self.stage_num += 1
        auto_config_text_file = self.create_postprocess_auto_config_text_file()
        submitter_params = self.template.get_submitter_params(TemplateSection.POSTPROCESS)
        submitter_params["hpc_config"] = _create_hpc_config(submitter_params)
        
        auto_config_py = self.get_auto_config_python_file()
        auto_config_command = f"python {auto_config_py} {auto_config_text_file}"
        config_params = self.template.get_config_params(TemplateSection.POSTPROCESS)
        stage = PipelineStage(
            auto_config_cmd=auto_config_command,
            config_file=config_params["config_file"],
            stage_num=self.stage_num,
            submitter_params=submitter_params
        )
        return stage


def _create_hpc_config(submitter_params):
    # TODO: this should be solved in JADE.
    hpc_config = submitter_params["hpc_config"]
    if isinstance(hpc_config, str):
        return HpcConfig(**load_data(hpc_config))

    if submitter_params["hpc_config"]["hpc_type"] == "local":
        return HpcConfig(hpc_type=HpcType.LOCAL, hpc=LocalHpcConfig())

    raise Exception(f"Unknown hpc config: {hpc_config}")



"""Computes hosting capacity"""

import csv
import itertools
import logging
import os
from concurrent.futures import ProcessPoolExecutor

import click
import pandas as pd
import numpy as np

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.jobs.results_aggregator import ResultsAggregator
from jade.loggers import setup_logging
from jade.utils.utils import get_cli_string, load_data

from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration


@click.command()
@click.option(
    "-o", "--jade-runtime-output",
    required=True,
    help="jade runtime output directory",
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Enable debug logging",
)
def compute_snapshot_hosting_capacity(jade_runtime_output, verbose=False):
    """Merge the pydss simulation results into one CSV file."""
    level = logging.DEBUG if verbose else logging.INFO
    logger = setup_logging(__name__, None, console_level=level, packages=["disco"])
    logger.info("Run compute_snapshot_hosting_capacity")
    print(get_cli_string())

    output_config_path = os.path.join(jade_runtime_output, CONFIG_FILE)
    config = PyDssConfiguration.deserialize(output_config_path)
    power_flow_job_names = set([
        x.name for x in config.iter_jobs() if x.extension == "pydss_simulation"
    ])
    job_pp_path = os.path.join(jade_runtime_output, JOBS_OUTPUT_DIR)
    all_pp_results = {}

    # build list of only successful results
    job_names = [
        x.name for x in ResultsAggregator.list_results(jade_runtime_output)
        if x.return_code == 0 and x.name in power_flow_job_names
    ]

    # create list of single job path to pass to executor
    job_paths = itertools.repeat(job_pp_path, len(job_names))
    jobs_post_process_results = {}

    with ProcessPoolExecutor() as executor:
        jobs_post_process_results = executor.map(_get_job_post_process_results,
                                                 job_paths, job_names)

    for job in jobs_post_process_results:
        job_config = config.get_job(job.get("name"))
        if job_config.feeder not in all_pp_results:
            all_pp_results.update({job_config.feeder: list()})

        all_pp_results[job_config.feeder].append(job)

    for feeder in all_pp_results:
        # output to csv
        result_df = pd.DataFrame(all_pp_results[feeder],
                                 columns=all_pp_results[feeder][0].keys())

        out_file = f"{feeder}-snapshot-impact-analysis-batch-post-process.csv"
        filename = os.path.join(jade_runtime_output, JOBS_OUTPUT_DIR, out_file)

        result_df.sort_values("penetration", inplace=True)
        result_df.to_csv(filename, index=False)
        logger.info("Dumped aggregated results to %s", out_file)

        # determine hosting capacity: group by feeder, penetration and check for max pv kw that passes
        def get_hosting_capacity(result_df):
            hc = {}
            for fail_field in fail_flag_columns:
                hc[fail_field.replace("flag", "hosting_capacity")] = result_df.loc[result_df[fail_field] ==
                                                                                   False]['pv_kw'].max()
            hc_df = pd.DataFrame(hc, index=[0])
            hc_columns = hc_df.filter(regex='hosting_capacity').columns
            hc_df[hc_columns] = hc_df[hc_columns].fillna(0)
            return hc_df

        # individual violation flags are fail flags
        # combination flags (contain word pass) are pass flags
        flag_columns = result_df.filter(regex='flag').columns
        result_df[flag_columns] = result_df[flag_columns].dropna()
        # convert flag columns to boolean type
        result_df[flag_columns] = result_df[flag_columns].replace({'False': False, 'True': True})
        result_df[['pv_kw', 'pv_pmpp']] = result_df[['pv_kw', 'pv_pmpp']].replace('', np.nan)
        result_df[['pv_kw', 'pv_pmpp']] = result_df[['pv_kw', 'pv_pmpp']].astype(float)

        # convert pass flags to fail flags (so there is uniformity in all flags)
        pass_flags = result_df.filter(regex='pass').columns
        for flag in pass_flags:
            result_df[flag] = ~ result_df[flag]

        result_df.columns = result_df.columns.str.replace('_pass', '')
        fail_flag_columns = result_df.filter(regex='flag').columns

        hosting_capacity_df = result_df.groupby('feeder').apply(get_hosting_capacity)
        hosting_capacity_df = hosting_capacity_df.reset_index().drop(columns='level_1')

        hc_file = f"{feeder}-snapshot-hosting-capacity-batch-post-process.csv"
        filename = os.path.join(jade_runtime_output, JOBS_OUTPUT_DIR, hc_file)
        hosting_capacity_df.to_csv(filename)


def _get_job_post_process_results(job_output_dir, job_name):
    """Get job post-process-results data

    Parameters
    ----------
    job_output_dir : str
    job_name : str

    Returns
    -------
    dict
        Dictionary of job name and all data from the job post-process results
    """
    results_file = os.path.join(
        job_output_dir,
        job_name,
        "snapshot-impact-analysis-job-post-process.csv",
    )

    with open(results_file) as f_in:
        results = []
        reader = csv.DictReader(f_in)
        results = list(reader)
        assert len(results) == 1, str(results)
        result = results[0]

    # add job name to results
    result["name"] = job_name
    return result



import logging
import sys
import time
from pathlib import Path

import click
import pandas as pd
import numpy as np

from jade.loggers import setup_logging
import disco


LOGGER_NAME = "CBA"
logger = logging.getLogger(LOGGER_NAME)

COMM_LOAD = "Loads__Powers__commercial (kWh)"
RES_LOAD = "Loads__Powers__residential (kWh)"
PVCOMM_OUT = "PVSystems__Powers__commercial (kWh)"
PVRES_OUT = "PVSystems__Powers__residential (kWh)"
SUB_POWER = "Circuits__TotalPower (kWh)"
SUB_LOSS = "Circuits__Losses (kWh)"
PVCOMM_CURT = "PVSystems__Curtailment__commercial (kWh)"
PVRES_CURT = "PVSystems__Curtailment__residential (kWh)"
CR1 = "Commercial power ($)"
CR2 = "Residential power ($)"
CR3 = "Commercial PV ($)"
CR4 = "Residential PV ($)"
CR5 = "Substation power ($)"
CR6 = "Substation losses ($)"
CR7 = "Commercial curtailment ($)"
CR8 = "Residential curtailment ($)"

COL_LIST = [
    COMM_LOAD,
    RES_LOAD,
    PVCOMM_OUT,
    PVRES_OUT,
    SUB_POWER,
    SUB_LOSS,
    PVCOMM_CURT,
    PVRES_CURT,
]
LABEL_COLUMNS = [
    "scenario",
    "name",
    "substation",
    "feeder",
    "placement",
    "sample",
    "penetration_level",
]
COSTS_PER_HOUR_FILE = Path(disco.__path__[0]) / "analysis" / "cba_costs_per_hour.csv"


def run(powers_file: Path, costs_file: Path, output_file: Path, skipna: bool):
    """Compute cost summaries."""
    start = time.time()
    powers_df = pd.read_csv(
        powers_file,
        dtype={"sample": np.float64, "penetration_level": np.float64, "placement": str},
    )
    costs_df = pd.read_csv(costs_file)
    raw_costs = []

    for (name, scenario), df in powers_df.groupby(by=["name", "scenario"]):
        if len(df) != 8760:
            raise Exception(f"Length of DataFrame is not 8760: {name}/{scenario}: {len(df)}")

        cost_list = []
        for col in LABEL_COLUMNS:
            unique_vals = df[col].unique()
            if len(unique_vals) != 1:
                raise Exception(f"Column does not have one unique value: {col}: {unique_vals}")
            cost_list.append(df[col].values[0])
        for col in COL_LIST:
            null_count = df[col].isnull().sum()
            if null_count > 0:
                logger.warning("Column %s has %s null values. skipna=%s", col, null_count, skipna)
            result = (df[col] * costs_df[col].values).sum(skipna=skipna)
            cost_list.append(result)
        raw_costs.append(cost_list)

    columns = LABEL_COLUMNS + [CR1, CR2, CR3, CR4, CR5, CR6, CR7, CR8]
    results_df = pd.DataFrame(raw_costs, columns=columns)
    results_df.to_csv(output_file, index=False)
    logger.info(f"Created %s. Duration = %s seconds", output_file, time.time() - start)


@click.command()
@click.argument("powers-file", type=click.Path(exists=True))
@click.option(
    "-c",
    "--costs-file",
    type=click.Path(exists=True),
    default=COSTS_PER_HOUR_FILE,
    show_default=True,
    help="Path to file containing hourly cost information",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(),
    default="cba_results.csv",
    show_default=True,
    help="Output file for results",
    callback=lambda _, __, z: Path(z),
)
@click.option(
    "--skipna/--no-skipna",
    is_flag=True,
    default=False,
    show_default=True,
    help="Skip null values when summing costs.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite output file if it already exists.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def compute_cba(powers_file, costs_file, output_file, skipna, force, verbose):
    """Compute Cost Benefit Analysis metrics."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(LOGGER_NAME, None, console_level=level, packages=["disco"])
    if output_file.exists() and not force:
        logger.error(
            "%s already exists. Pass --force to overwrite or choose a custom name.", output_file
        )
        sys.exit(1)
    run(powers_file, costs_file, output_file, skipna)



"""Defines Input objects."""

import json


class Input:
    """Input class used to store analysis input values"""

    def __init__(self, name, input_type, default, value=None):
        self.name = name
        self.input_type = input_type
        self.default = default
        self.value = value

    def serialize(self):
        """Serializes input to dict"""
        return dict({
            'name': self.name,
            'input_type': self.input_type.get_type_name(),
            'default': self.default,
            'value': self.value
        })

    @property
    def current_value(self):
        """Get Input current value (converted if necessary)"""
        input_type = self.input_type.get_type_name()
        if input_type == 'percent':
            return self.value / 100

        return self.value

    def to_json(self):
        """Output variables to json string"""

        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)



import logging

import click

from jade.exceptions import InvalidParameter
from jade.utils.utils import load_data
from disco.models.factory import list_model_classes

logger = logging.getLogger(__name__)


@click.group()
def simulation_models():
    """Commands to operate on simulation models"""


def get_model_class(ctx, param, value):
    for cls in list_model_classes():
        if cls.__name__ == value:
            return cls
    raise InvalidParameter(f"invalid model: {value}")


@click.command(name="list")
def list_models():
    """List all simulation models."""
    print("{}".format(" ".join([x.__name__ for x in list_model_classes()])))


@click.command()
@click.argument("model", callback=get_model_class)
def show_example(model):
    """Show an example job definition for a simulation model."""
    print(model.example_json())


@click.command()
@click.argument("model", callback=get_model_class)
def show_schema(model):
    """Show the schema for a simulation model."""
    print(model.schema_json())


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.argument("model", callback=get_model_class)
def validate_file(filename, model):
    """Validate the JSON inputs config file."""
    data = load_data(filename)
    if isinstance(data, dict):
        model.validate(data)
    else:
        for item in data:
            model.validate(item)
    print("Success!")


simulation_models.add_command(list_models)
simulation_models.add_command(show_example)
simulation_models.add_command(show_schema)
simulation_models.add_command(validate_file)




import os

from jade.utils.utils import load_data
from disco.distribution.distribution_inputs import DistributionInputs
from disco.extensions.upgrade_simulation.upgrade_parameters import UpgradeParameters


class UpgradeInputs(DistributionInputs):
    
    def __init__(self, base_directory):
        super().__init__(base_directory)

    def _parse_config_files(self):
        filename = os.path.join(self._base, self._CONFIG_FILE)
        data = load_data(filename)
        for job_data in data:
            job = UpgradeParameters(**job_data)
            assert job.name not in self._parameters
            self._parameters[job.name] = job



import logging
import os
import sys

import pandas as pd
import matplotlib.pyplot as plt 

logger = logging.getLogger(__name__)


def plot_voltage(output_dir, scenario=None):
    """
    Create 3 plots based on the first feeder,
    1. compare voltage primary and secondary.
    2. compare pf1 and volt-var
    3. heatmap for max and min hosting capacity

    Parameters
    ----------
    output_dir : str | pathlib.Path
        The output directory that contains the metrics and hosting capacity results.
    scenario : str
        The scenario name of simulation, default None.
    """
    voltage_metrics_table = os.path.join(output_dir, "voltage_metrics_table.csv")
    voltage_metrics = pd.read_csv(voltage_metrics_table)
    feeder_example = voltage_metrics['feeder'].unique()[0]
    voltage_metrics = voltage_metrics[voltage_metrics['feeder']==feeder_example]

    fig, ax = plt.subplots(figsize=(8,8))
    ax.scatter(
        voltage_metrics[voltage_metrics['node_type']=='primaries']['penetration_level'],
        voltage_metrics[voltage_metrics['node_type']=='primaries']["max_voltage"],
        facecolors='none',
        edgecolors='C0',
        label="primary"
    )
    ax.scatter(
        voltage_metrics[voltage_metrics['node_type']=='secondaries']['penetration_level'],
        voltage_metrics[voltage_metrics['node_type']=='secondaries']["max_voltage"],
        facecolors='none',
        edgecolors='C1',
        label="secondary"
    )
    ax.legend()
    ax.set_title(feeder_example)
    ax.set_xlabel("Penetration level")
    ax.set_ylabel("max_voltage (pu)")
    fig.savefig(os.path.join(output_dir,"max_voltage_pri_sec.png"))

    fig, ax = plt.subplots(figsize=(8,8))
    ax.scatter(
        voltage_metrics[voltage_metrics['scenario']=='pf1']['penetration_level'],
        voltage_metrics[voltage_metrics['scenario']=='pf1']["max_voltage"],
        facecolors='none',
        edgecolors='C0',
        label="base_case:pf1"
    )
    ax.scatter(
        voltage_metrics[voltage_metrics['scenario']=='control_mode']['penetration_level'],
        voltage_metrics[voltage_metrics['scenario']=='control_mode']["max_voltage"],
        facecolors='none',
        edgecolors='C1',
        label="control_mode:volt-var"
    )
    ax.legend()
    ax.set_title(feeder_example)
    ax.set_xlabel("Penetration level")
    ax.set_ylabel("max_voltage (pu)")
    fig.savefig(os.path.join(output_dir,"max_voltage_pf1_voltvar.png"))
    logger.info("Voltage plot created.")


def plot_hc(output_dir, scenario):
    """
    Plot hosting capacity heatmap for all feeders,

    Parameters
    ----------
    output_dir : str | pathlib.Path
        The output directory that contains the metrics and hosting capacity results.
    scenario : str
        The scenario name of simulation.
    """
    overall_file = os.path.join(output_dir, f"hosting_capacity_overall__{scenario}.json")
    if not os.path.exists(overall_file):
        logger.error("Overall hosting capacity JSON file does not exist, please check your scenario.")
        sys.exit(1)

    overall = pd.read_json(overall_file).transpose()

    _, ax = plt.subplots(figsize=(8,8))
    y_pos = overall.index.to_list()
    ax.barh(
        y_pos,
        overall['min_hc_pct'],
        label="no violation",
        color='limegreen'
    )
    ax.barh(
        y_pos,
        overall['max_hc_pct']-overall['min_hc_pct'],
        left=overall['min_hc_pct'],
        label="some violation",
        color='gold'
    )
    ax.barh(
        y_pos,
        200-overall['max_hc_pct'],
        left=overall['max_hc_pct'],
        label="violation",
        color='tomato'
    )
    ax.set_title(f"HCA heatmap: {scenario}")
    ax.set_xlabel("Penetration level (%)")
    ax.legend(ncol=3)
    plt.savefig(os.path.join(output_dir,f"hca__{scenario}.png"))
    
    logger.info("Hosting capacity plot created.")



"""Factory functions for creating data generators."""

from jade.exceptions import InvalidParameter
from jade.utils.timing_utils import timed_info
from jade.utils.utils import load_data
from .open_dss_generator import OpenDssGenerator


@timed_info
def read_config_data(config_file):
    """Return a model generator based on the config file."""
    data = load_data(config_file)

    if data["type"] == "GemModel":
        cls = OpenDssGenerator
    else:
        raise InvalidParameter(f"unsupported model type {data['type']}")

    generator = cls(data)
    return generator



"""Make cost benefit analysis summary tables for all jobs in a batch."""

import csv
import itertools
import json
import logging
from collections import defaultdict, namedtuple
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

import click
import pandas as pd
import toml

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.jobs.results_aggregator import ResultsAggregator
from jade.utils.subprocess_manager import check_run_command
from jade.utils.utils import load_data

from PyDSS.pydss_results import PyDssResults

from disco.pipelines.utils import ensure_jade_pipeline_output_dir


JobInfo = namedtuple(
    "JobInfo", ["name", "substation", "feeder", "placement", "sample", "penetration_level"]
)
REQUIRED_RESOLUTION = pd.Timedelta(hours=1)

logger = logging.getLogger(__name__)


def parse_batch_results(output_dir):
    """Parse the results from all jobs in a JADE output directory."""
    output_path = Path(output_dir)
    config = create_config_from_file(output_path / CONFIG_FILE)
    jobs = []
    results = ResultsAggregator.list_results(output_dir)
    result_lookup = {x.name: x for x in results}
    for job in config.iter_pydss_simulation_jobs():
        if job.name not in result_lookup:
            logger.info("Skip missing job %s", job.name)
            continue
        if result_lookup[job.name].return_code != 0:
            logger.info("Skip failed job %s", job.name)
            continue
        jobs.append(job)

    powers_tables = []
    capacitor_table = []
    reg_control_change_table = []
    load_sum_group_files = set()
    pv_sum_group_files = set()

    # This will create flattened tables for each metric across jobs and PyDSS scenarios
    # within eah job.
    # Every table contains job name, substation, feeder, etc., and scenario name, as
    # well as the metrics.
    with ProcessPoolExecutor() as executor:
        for result in executor.map(parse_job_results, jobs, itertools.repeat(output_path)):
            powers_tables.append(result[0])
            capacitor_table += result[1]
            reg_control_change_table += result[2]
            load_sum_group_files.update(result[3])
            pv_sum_group_files.update(result[4])

    df = pd.concat(powers_tables)
    powers_file = output_path / "powers_table.csv"
    df.to_csv(powers_file)
    logger.info("Wrote power data to %s", output_path / "powers_table.csv")
    serialize_table(capacitor_table, output_path / "capacitor_table.csv")
    serialize_table(
        reg_control_change_table, output_path / "reg_control_tap_value_change_table.csv"
    )
    loads = make_customer_type_table(load_sum_group_files, "Load")
    serialize_table(loads, output_path / "load_customer_types.csv")
    pvs = make_customer_type_table(pv_sum_group_files, "PVSystem")
    serialize_table(pvs, output_path / "pv_system_customer_types.csv")
    check_run_command(f"disco-internal cba-post-process {powers_file}")


def parse_job_results(job, output_path):
    """Return the tables for a single job."""
    job_path = output_path / JOBS_OUTPUT_DIR / job.name / "pydss_project"
    deployment = job.model.deployment
    if job.model.is_base_case:
        job_info = JobInfo(
            name=job.name,
            substation=deployment.substation,
            feeder=deployment.feeder,
            placement="",
            sample="",
            penetration_level="",
        )
    else:
        job_info = JobInfo(
            name=job.name,
            substation=deployment.substation,
            feeder=deployment.feeder,
            placement=deployment.project_data["placement"],
            sample=deployment.project_data["sample"],
            penetration_level=deployment.project_data["penetration_level"],
        )
    results = PyDssResults(job_path)
    powers_table = get_powers_table(results, job_info)
    capacitor_table = get_capacitor_table(results, job_info)
    reg_control_table = get_reg_control_table(results, job_info)
    load_sum_group_files, pv_sum_group_files = get_sum_group_files(results)

    return (
        powers_table,
        capacitor_table,
        reg_control_table,
        load_sum_group_files,
        pv_sum_group_files,
    )


def get_powers_table(results: PyDssResults, job_info: JobInfo):
    """Return pd.DataFrame containing all power data."""
    dfs = []
    column_order = list(job_info._asdict().keys()) + [
        "scenario",
        "Loads__Powers__commercial (kWh)",
        "Loads__Powers__residential (kWh)",
        "PVSystems__Powers__commercial (kWh)",
        "PVSystems__Powers__residential (kWh)",
        "Circuits__TotalPower (kWh)",
        "Circuits__Losses (kWh)",
    ]
    for scenario in results.scenarios:
        main_df = pd.DataFrame()
        missing = []
        for customer in ("commercial", "residential"):
            for elem_class in ("Loads", "PVSystems"):
                column = f"{elem_class}__Powers__{customer} (kWh)"
                if elem_class not in scenario.list_element_classes():
                    # Base case doesn't have PVSystems.
                    missing.append(column)
                    continue
                prop = f"Powers__{customer}"
                if prop in scenario.list_summed_element_time_series_properties(elem_class):
                    df = scenario.get_summed_element_dataframe(elem_class, prop)
                    if len(main_df) == 0:
                        main_df.index = df.index
                    # Exclude neutral phase.
                    cols = [col for col in df.columns if "__N" not in col]
                    assert len(df.columns) == 1
                    series = df.iloc[:, 0]
                    if elem_class == "Loads":
                        data = series.values
                    else:
                        data = series.apply(get_pv_power_value).values
                    main_df[column] = data
                else:
                    # Not all jobs will have commercial and residential.
                    missing.append(column)
        for column in missing:
            main_df[column] = 0.0

        df = scenario.get_full_dataframe("Circuits", "TotalPower")
        assert len(df.columns) == 1
        main_df["Circuits__TotalPower (kWh)"] = [x.real for x in df[df.columns[0]]]
        # OpenDSS returns this in Watts.
        df = scenario.get_full_dataframe("Circuits", "Losses") / 1000
        assert len(df.columns) == 1
        main_df["Circuits__Losses (kWh)"] = [x.real for x in df[df.columns[0]]]
        if main_df.index[1] - main_df.index[0] == REQUIRED_RESOLUTION:
            df = main_df
        else:
            df = main_df.resample(REQUIRED_RESOLUTION).mean()

        for field, val in job_info._asdict().items():
            df[field] = val
        df["scenario"] = scenario.name
        df = df[column_order]
        dfs.append(df)

    return pd.concat(dfs)


def get_pv_power_value(val):
    if val > 0.0:
        if val < 0.001:  # 1 Watt
            val = 0.0
        else:
            logger.warning("Unexpected PVSystem power value: %s", val)
            val *= -1
    else:
        val *= -1
    return val


def get_capacitor_table(results: PyDssResults, job_info: JobInfo):
    """Return capacitor state changes for all scenarios."""
    capacitor_table = []
    data = json.loads(results.read_file("Reports/capacitor_state_changes.json"))
    for scenario in data["scenarios"]:
        for capacitor in scenario["capacitors"]:
            row = job_info._asdict()
            row["device_name"] = capacitor["name"]
            row["action_count"] = capacitor["change_count"]
            capacitor_table.append(row)
    return capacitor_table


def get_reg_control_table(results: PyDssResults, job_info: JobInfo):
    """Return reg_control state changes for all scenarios."""
    reg_control_table = []
    data = json.loads(results.read_file("Reports/reg_control_tap_value_changes.json"))
    for scenario in data["scenarios"]:
        for reg_control in scenario["reg_controls"]:
            row = job_info._asdict()
            row["device_name"] = reg_control["name"]
            row["action_count"] = reg_control["change_count"]
            reg_control_table.append(row)
    return reg_control_table


def get_sum_group_files(results: PyDssResults):
    """Return a set of filenames defining sum_groups and Load/PV customer types."""
    load_sum_group_files = set()
    pv_sum_group_files = set()
    for scenario in results.scenarios:
        data = toml.loads(
            scenario.read_file(f"Scenarios/{scenario.name}/ExportLists/Exports.toml")
        )
        load_sum_group_files.add(data["Loads"]["Powers"]["sum_groups_file"])
        pv_sum_group_files.add(data["PVSystems"]["Powers"]["sum_groups_file"])
    return load_sum_group_files, pv_sum_group_files


def make_customer_type_table(sum_group_files, element_type):
    """Return a dict of customer type to set of element names."""
    elements = defaultdict(set)
    for filename in sum_group_files:
        data = load_data(filename)
        for group in data["sum_groups"]:
            elements[group["name"]].update(set(group["elements"]))

    table = []
    for group, element_names in elements.items():
        for name in element_names:
            table.append({"customer_type": group, "element_type": element_type, "name": name})
    return table


def serialize_table(table, filename):
    """Serialize a list of dictionaries to a CSV file."""
    with open(filename, "w") as f:
        if table:
            fields = table[0].keys()
            writer = csv.DictWriter(f, fieldnames=fields)
            writer.writeheader()
            writer.writerows(table)
        else:
            f.write('""\n')
        logger.info("Wrote %s", filename)


@click.command()
@click.argument("output_dir")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def make_cba_tables(output_dir, verbose):
    """Make cost benefit analysis summary tables for all jobs in a batch."""
    level = logging.DEBUG if verbose else logging.INFO
    output_dir = Path(ensure_jade_pipeline_output_dir(output_dir))
    log_file = output_dir / "make_cba_tables.log"
    setup_logging(__name__, log_file, file_level=level, console_level=level, packages=["disco"])
    parse_batch_results(output_dir)



import os
import logging
import pandas as pd
import numpy as np

from jade.utils.timing_utils import track_timing, Timer
from jade.utils.utils import load_data, dump_data

from .common_functions import create_overall_output_file, convert_dict_nan_to_none, summarize_upgrades_outputs, \
    convert_length_units
from disco import timer_stats_collector
from disco.utils.custom_encoders import ExtendedJSONEncoder
from disco.models.upgrade_cost_analysis_generic_input_model import load_cost_database
from disco.models.upgrade_cost_analysis_generic_output_model import AllUpgradesTechnicalResultModel, \
    AllEquipmentUpgradeCostsResultModel, EquipmentTypeUpgradeCostsResultModel, CapacitorControllerResultType, \
        VoltageRegulatorResultType, TotalUpgradeCostsResultModel, AllUpgradesCostResultSummaryModel
    
logger = logging.getLogger(__name__)


@track_timing(timer_stats_collector)
def compute_all_costs(
    job_name,
    output_json_thermal_upgrades_filepath,
    output_json_voltage_upgrades_filepath,
    cost_database_filepath,
    output_equipment_upgrade_costs_filepath,
    output_total_upgrade_costs_filepath,
    overall_output_summary_filepath,
    feeder_stats_json_file,
):
    # upgrades files
    all_upgrades = {}
    all_upgrades.update(load_data(output_json_voltage_upgrades_filepath))
    all_upgrades.update(load_data(output_json_thermal_upgrades_filepath))
    # validate upgrades details for thermal and voltage, using pydantic models
    m = AllUpgradesTechnicalResultModel(**all_upgrades)
    xfmr_upgrades_df = pd.DataFrame(m.dict(by_alias=True)["transformer"])
    line_upgrades_df = pd.DataFrame(m.dict(by_alias=True)["line"])
    voltage_upgrades_df = pd.DataFrame(m.dict(by_alias=True)["voltage"])
    
    (
        xfmr_cost_database,
        line_cost_database,
        controls_cost_database,
        voltage_regulators_cost_database,
        misc_database,
    ) = load_cost_database(cost_database_filepath)
    output_columns = list(EquipmentTypeUpgradeCostsResultModel.schema(True).get("properties").keys())
    # reformat data
    if not xfmr_upgrades_df.empty:
        xfmr_upgrades_df, xfmr_cost_database = reformat_xfmr_files(
            xfmr_upgrades_df=xfmr_upgrades_df, xfmr_cost_database=xfmr_cost_database)
        # compute thermal upgrade costs
        xfmr_cost_df = compute_transformer_costs(xfmr_upgrades_df=xfmr_upgrades_df, xfmr_cost_database=xfmr_cost_database,
                                                 misc_database=misc_database)
        xfmr_cost_df["name"] = job_name
    else:
        xfmr_cost_df = pd.DataFrame(columns=output_columns).astype({"count": int, "total_cost_usd": float})

    if not line_upgrades_df.empty:
        line_upgrades_df, line_cost_database = reformat_line_files(
            line_upgrades_df=line_upgrades_df, line_cost_database=line_cost_database)
        line_cost_df = compute_line_costs(line_upgrades_df=line_upgrades_df, line_cost_database=line_cost_database)
        line_cost_df["name"] = job_name
    else:
        line_cost_df = pd.DataFrame(columns=output_columns).astype({"count": int, "total_cost_usd": float})

    thermal_cost_df = pd.concat([xfmr_cost_df, line_cost_df])

    if not voltage_upgrades_df.empty:
        # compute voltage upgrade costs
        cap_cost_df = compute_capcontrol_cost(voltage_upgrades_df=voltage_upgrades_df,
                                            controls_cost_database=controls_cost_database)
        reg_cost_df = compute_voltage_regcontrol_cost(voltage_upgrades_df=voltage_upgrades_df,
                                                    vreg_control_cost_database=controls_cost_database, 
                                                    vreg_xfmr_cost_database=voltage_regulators_cost_database, xfmr_cost_database=xfmr_cost_database)
        voltage_cost_df = pd.concat([cap_cost_df, reg_cost_df])
        voltage_cost_df["name"] = job_name
    else:
        voltage_cost_df = pd.DataFrame(columns=output_columns).astype({"count": int, "total_cost_usd": float})

    voltage_cost_df = voltage_cost_df.loc[voltage_cost_df["count"] != 0]
    thermal_cost_df = thermal_cost_df.loc[thermal_cost_df["count"] != 0]
    total_cost_df = get_total_costs(thermal_cost_df, voltage_cost_df)
    equipment_costs = AllEquipmentUpgradeCostsResultModel(thermal=thermal_cost_df.to_dict('records'), voltage=voltage_cost_df.to_dict('records'))
    dump_data(convert_dict_nan_to_none(equipment_costs.dict(by_alias=True)), 
              output_equipment_upgrade_costs_filepath, indent=2, cls=ExtendedJSONEncoder, allow_nan=False)
    total_cost_df["name"] = job_name
    m = [TotalUpgradeCostsResultModel(**x) for x in total_cost_df.to_dict(orient="records")]
    total_costs_per_equipment = convert_dict_nan_to_none({"costs_per_equipment": total_cost_df.to_dict('records')})
    feeder_stats = load_data(feeder_stats_json_file)
    output_summary = create_overall_output_file(upgrades_dict={"transformer": xfmr_upgrades_df, "line": line_upgrades_df, "voltage": voltage_upgrades_df},
                                                costs_dict={"thermal": thermal_cost_df, "voltage": voltage_cost_df}, feeder_stats=feeder_stats, job_name=job_name)
    output_summary_model = AllUpgradesCostResultSummaryModel(equipment=output_summary.to_dict(orient="records"))
    output_summary = convert_dict_nan_to_none(output_summary_model.dict(by_alias=True))
    if os.path.exists(overall_output_summary_filepath):
        overall_outputs = load_data(overall_output_summary_filepath)
        overall_outputs.update(total_costs_per_equipment)
    else:
        overall_outputs = total_costs_per_equipment
    overall_outputs.update(output_summary)  
    overall_outputs.update(summarize_upgrades_outputs(overall_outputs, job_name=job_name))
    desired_order_list = ["results", "costs_per_equipment", "violation_summary", "equipment"]
    reordered_dict = {k: overall_outputs[k] for k in desired_order_list}
    dump_data(reordered_dict, overall_output_summary_filepath, indent=2, cls=ExtendedJSONEncoder, allow_nan=False) 
    

def compute_transformer_costs(xfmr_upgrades_df, xfmr_cost_database, **kwargs):
    """This function computes the transformer costs.
    -Unit equipment cost for new_parallel and "upgrade" transformers are the same in the database.
    The difference would be the fixed costs added (if present in misc_database)
    -For transformers that are of action "upgrade":
    transformers of old rating are removed, and that of upgraded rating is added.
    -For transformers that are of action "new_parallel":
    A new transformer is placed in parallel with the existing transformer
    -These are the properties considered while choosing unit cost:
    ["rated_kVA", "phases", "primary_kV", "secondary_kV", "primary_connection_type", "secondary_connection_type",
    "num_windings"]
    -For a given transformer, if unit cost pertaining to these properties is not available,
    then the closest "rated_kVA" unit cost is chosen.
    User can decide if they want to choose backup based on another property.
    Parameters
    ----------
    xfmr_upgrades_df
    xfmr_cost_database
    kwargs

    Returns
    -------

    """
    output_cost_field = "total_cost_usd"
    output_count_field = "count"
    deciding_columns = ["rated_kVA", "phases", "primary_kV", "secondary_kV", "primary_connection_type",
                        "secondary_connection_type", "num_windings"]
    output_columns_list = ["type", output_count_field, output_cost_field, "comment", "equipment_parameters"]
    backup_deciding_property = kwargs.get("backup_deciding_property", "rated_kVA")
    misc_database = kwargs.get("misc_database", None)
    # choose which properties are to be saved
    upgrade_type_list = ["upgrade", "new_parallel"]
    added_xfmr_df = xfmr_upgrades_df.loc[(xfmr_upgrades_df["upgrade_type"].isin(upgrade_type_list)) & (xfmr_upgrades_df["action"] == "add")]
    computed_cost = []
    for index, row in added_xfmr_df.iterrows():
        unit_cost = xfmr_cost_database.loc[(xfmr_cost_database["rated_kVA"] == row["rated_kVA"]) &
                                           (xfmr_cost_database["phases"] == row["phases"]) &
                                           (xfmr_cost_database["primary_kV"] == row["primary_kV"]) &
                                           (xfmr_cost_database["secondary_kV"] == row["secondary_kV"]) &
                                           (xfmr_cost_database["primary_connection_type"] == row["primary_connection_type"]) &
                                           (xfmr_cost_database["secondary_connection_type"] == row["secondary_connection_type"]) &
                                           (xfmr_cost_database["num_windings"] == row["num_windings"])
                                           ]["cost"]
        params_dict = dict(row[['final_equipment_name'] + deciding_columns])
        row["equipment_parameters"] = params_dict
        row["type"] = "Transformer"
        if len(unit_cost) > 0:  # if there are more than one rows, the first one is chosen in random
            unit_cost = unit_cost.values[0]
            row[output_cost_field] = unit_cost
            row["comment"] = ""
            row[output_count_field] = 1
        else:  # if costs are not present for this transformer, then choose closest rated_kVA
            # (or whatever backup deciding property is passed) (ignore other properties)
            closest = xfmr_cost_database.loc[abs(xfmr_cost_database[backup_deciding_property] -
                                                 row[backup_deciding_property]).idxmin()].copy()
            row[output_cost_field] = closest["cost"]
            row[output_count_field] = 1
            for key, value in closest.items():  # make it json serializable
                if isinstance(value, np.int64):
                    closest[key] = int(value)
            comment_string = {"text": f"Transformer {row['final_equipment_name']}: Exact cost not available. " \
                                      f"Unit cost for transformer with these parameters used " \
                                      f"(based on closest {backup_deciding_property}", 
                             "params": dict(closest)}
            logger.debug(comment_string)
            row["comment"] = comment_string
        # add transformer fixed costs, if given in database. (depending on upgrade type)
        if (misc_database is not None) and (not misc_database.empty):
            misc_xfmr_fields = {"replace": "Replace transformer (fixed cost)",
                                "new": "Add new transformer (fixed cost)"}
            # if equipment is upgraded, and misc database contains fixed cost for replacing xfmr
            if (row["upgrade_type"].lower() == "upgrade") and \
                    ((misc_database["description"] == misc_xfmr_fields["replace"]).any()):
                field_name = misc_xfmr_fields["replace"]
                fixed_cost = misc_database.loc[misc_database["description"] == field_name]
            # if equipment is new, and misc database contains fixed cost for adding new xfmr
            elif row["upgrade_type"].lower() == "new_parallel" and \
                    ((misc_database["description"] == misc_xfmr_fields["new"]).any()):
                field_name = misc_xfmr_fields["new"]
                fixed_cost = misc_database.loc[misc_database["description"] == field_name]
            else:
                fixed_cost = pd.DataFrame()
            if not fixed_cost.empty:
                row[output_cost_field] += misc_database.loc[misc_database["description"]
                                                            == field_name]["total_cost"].values[0]

        computed_cost.append(row[output_columns_list])
    xfmr_cost_df = pd.DataFrame(computed_cost)
    return xfmr_cost_df


def reformat_xfmr_files(xfmr_upgrades_df, xfmr_cost_database):
    """This function renames, reformats transformer upgrades dataframe to match cost database columns

    Parameters
    ----------
    xfmr_upgrades_df
    xfmr_cost_database

    Returns
    -------

    """
    xfmr_upgrades_df.rename(columns={"kVA": "rated_kVA", "windings": "num_windings"}, inplace=True)
    xfmr_upgrades_df["rated_kVA"] = xfmr_upgrades_df["rated_kVA"].astype(float)
    xfmr_upgrades_df["num_windings"] = xfmr_upgrades_df["num_windings"].astype(int)
    xfmr_upgrades_df["phases"] = xfmr_upgrades_df["phases"].astype(int)
    xfmr_upgrades_df["primary_kV"] = xfmr_upgrades_df["kVs"].str[0].astype(float)
    xfmr_upgrades_df["secondary_kV"] = xfmr_upgrades_df["kVs"].str[-1].astype(float)
    xfmr_upgrades_df["primary_connection_type"] = xfmr_upgrades_df["conns"].str[0]
    xfmr_upgrades_df["secondary_connection_type"] = xfmr_upgrades_df["conns"].str[-1]

    xfmr_cost_database["rated_kVA"] = xfmr_cost_database["rated_kVA"].astype(float)
    xfmr_cost_database["num_windings"] = xfmr_cost_database["num_windings"].astype(int)
    xfmr_cost_database["phases"] = xfmr_cost_database["phases"].astype(int)
    xfmr_cost_database["primary_kV"] = xfmr_cost_database["primary_kV"].astype(float)
    xfmr_cost_database["secondary_kV"] = xfmr_cost_database["secondary_kV"].astype(float)
    xfmr_cost_database["cost"] = xfmr_cost_database["cost"].astype(float)

    return xfmr_upgrades_df, xfmr_cost_database


def reformat_xfmr_upgrades_file(xfmr_upgrades_df):
    xfmr_upgrades_df.rename(columns={"kVA": "rated_kVA", "windings": "num_windings"}, inplace=True)
    xfmr_upgrades_df["rated_kVA"] = xfmr_upgrades_df["rated_kVA"].astype(float)
    xfmr_upgrades_df["num_windings"] = xfmr_upgrades_df["num_windings"].astype(int)
    xfmr_upgrades_df["phases"] = xfmr_upgrades_df["phases"].astype(int)
    xfmr_upgrades_df["primary_kV"] = xfmr_upgrades_df["kVs"].str[0].astype(float)
    xfmr_upgrades_df["secondary_kV"] = xfmr_upgrades_df["kVs"].str[-1].astype(float)
    xfmr_upgrades_df["primary_connection_type"] = xfmr_upgrades_df["conns"].str[0]
    xfmr_upgrades_df["secondary_connection_type"] = xfmr_upgrades_df["conns"].str[-1]
    return xfmr_upgrades_df


def compute_line_costs(line_upgrades_df, line_cost_database, **kwargs):
    """This function computes the line costs.
    -Unit equipment cost for new_parallel and "upgrade" line are the not same in the database.
    There are different costs given for reconductored and new lines
    -For lines that are of action "upgrade": "reconductored" line unit costs need to be used
    -For lines that are of action "new_parallel": "new" line unit costs need to be used
    -Upgraded lines and new lines run along existing circuit, so length is the same for both
    -These are the properties considered while choosing unit cost:
    ["phases", "voltage_kV", "ampere_rating", "line_placement", "description" (i.e. whether new or reconductored)]
    -For a given line, if unit cost pertaining to these properties is not available,
    then the closest "ampere_rating" unit cost is chosen.
    User can decide if they want to choose backup based on another property.
    For 3 phase lines, voltage_kV should be LN voltage (# TODO check if this is correct)

    Parameters
    ----------
    line_upgrades_df
    line_cost_database
    kwargs

    Returns
    -------

    """
    output_cost_field = "total_cost_usd"
    output_count_field = "count"
    deciding_columns = ["phases", "voltage_kV", "ampere_rating", "line_placement", "upgrade_type"]
    output_columns_list = ["type", output_count_field, output_cost_field, "comment", "equipment_parameters"]
    backup_deciding_property = kwargs.get("backup_deciding_property", "ampere_rating")
    # choose which properties are to be saved
    upgrade_type_list = ["upgrade", "new_parallel"]
    
    added_line_df = line_upgrades_df.loc[(line_upgrades_df["upgrade_type"].isin(upgrade_type_list)) & (line_upgrades_df["action"] == "add")]
    computed_cost = []
    for index, row in added_line_df.iterrows():
        if row["upgrade_type"] == "upgrade":
            upgrade_type = "reconductored_line"
        elif row["upgrade_type"] == "new_parallel":
            upgrade_type = "new_line"
        else:
            # if anything else, by default, use new_line prices
            upgrade_type = "new_line"
        row["upgrade_type"] = upgrade_type
        unit_cost = line_cost_database.loc[(line_cost_database["phases"] == row["phases"]) &
                                           (line_cost_database["voltage_kV"] == row["voltage_kV"]) &
                                           (line_cost_database["ampere_rating"] == row["ampere_rating"]) &
                                           (line_cost_database["line_placement"] == row["line_placement"]) &
                                           (line_cost_database["upgrade_type"] == upgrade_type)
                                           ]["cost_per_m"]
        # OpenDSS can output results in any of these lengths.
        # convert line length to metres
        line_length_m = convert_length_units(length=row["length"], unit_in=row["units"], unit_out="m")
        params_dict = dict(row[['final_equipment_name'] + deciding_columns])
        row["equipment_parameters"] = params_dict
        row["type"] = "Line"
        if len(unit_cost) > 0:  # if there are more than one rows, the first one is chosen in random
            unit_cost = unit_cost.values[0]
            row[output_cost_field] = unit_cost * line_length_m
            row[output_count_field] = 1
            row["comment"] = ""
        else:  # if costs are not present for this transformer, then choose closest ampere_rating
            # (or whatever backup deciding property is passed) (ignore other properties)
            closest = line_cost_database.loc[abs(line_cost_database[backup_deciding_property] -
                                                 row[backup_deciding_property]).idxmin()].copy()
            row[output_cost_field] = closest["cost_per_m"] * line_length_m
            for key, value in closest.items():  # make it json serializable
                if isinstance(value, np.int64):
                    closest[key] = int(value)
            comment_string = {"text": f"Line {row['final_equipment_name']}: Exact cost not available. " \
                                      f"Unit cost for line with these parameters used " \
                                      f"(based on closest {backup_deciding_property}.",
                             "params": dict(closest)}
            
            logger.debug(comment_string)
            row["comment"] = comment_string
            row[output_count_field] = 1
        computed_cost.append(row[output_columns_list])
    line_cost_df = pd.DataFrame(computed_cost)
    return line_cost_df


def reformat_line_files(line_upgrades_df, line_cost_database):
    """This function renames, reformats line upgrades dataframe to match cost database columns

    Parameters
    ----------
    line_upgrades_df
    line_cost_database

    Returns
    -------

    """
    line_upgrades_df.rename(columns={"normamps": "ampere_rating", "kV": "voltage_kV"}, inplace=True)
    line_upgrades_df["ampere_rating"] = line_upgrades_df["ampere_rating"].astype(float).round(2)
    line_upgrades_df["phases"] = line_upgrades_df["phases"].astype(int)
    line_upgrades_df["voltage_kV"] = line_upgrades_df["voltage_kV"].astype(float).round(2)
    # assign original equipment length to new equipment
    line_upgrades_df["length"] = line_upgrades_df.groupby("original_equipment_name")["length"].transform("first")

    line_cost_database["ampere_rating"] = line_cost_database["ampere_rating"].astype(float).round(2)
    line_cost_database["phases"] = line_cost_database["phases"].astype(int)
    line_cost_database["voltage_kV"] = line_cost_database["voltage_kV"].astype(float).round(2)
    line_cost_database["cost_per_m"] = line_cost_database["cost_per_m"].astype(float)
    return line_upgrades_df, line_cost_database


def compute_capcontrol_cost(voltage_upgrades_df, controls_cost_database, keyword="Capacitor"):
    """This function computes the capacitor controller related costs.
    Note we currently are not adding new capacitors to integrate PV.
    Considered here: new controllers, control setting changes

    Parameters
    ----------
    voltage_upgrades_df
    controls_cost_database

    Returns
    -------

    """
    output_cost_field = "total_cost_usd"
    output_count_field = "count"
    output_columns_list = ["type", output_count_field, output_cost_field, "comment"]

    type_rows = CapacitorControllerResultType.list_values()  # from enum
    capcontrol_upgrade_fields = {"add_new_cap_controller": "new_controller_added",
                         "change_cap_control": "controller_settings_modified"}
    cost_database_fields = {"add_new_cap_controller": "Add new capacitor controller",
                            "change_cap_control": "Change capacitor controller settings",
                            "replace_cap_controller": "Replace capacitor controller"
                            }
    empty_cap_cost_dict = {"type": type_rows,
                           "count": [0] * len(type_rows),  "total_cost_usd": [0] * len(type_rows)}
    zero_cost_df = pd.DataFrame.from_dict(empty_cap_cost_dict)

    if voltage_upgrades_df.empty:  # if there are no voltage upgrades
        return zero_cost_df
    cap_upgrades_df = voltage_upgrades_df.loc[voltage_upgrades_df['equipment_type'].str.contains(keyword)]
    if cap_upgrades_df.empty:  # if there are no capacitor control upgrades
        return zero_cost_df
    cap_cost = []
    # if there are new capacitor controller 
    count_new_controller = cap_upgrades_df[capcontrol_upgrade_fields["add_new_cap_controller"]].sum()
    unit_cost_new_controller = controls_cost_database.loc[controls_cost_database["type"]
                                                          == cost_database_fields["add_new_cap_controller"]]["cost"].values[0]
    total_cost_new_controller = count_new_controller * unit_cost_new_controller
    cap_cost.append( {"type": CapacitorControllerResultType.add_new_cap_controller.value,
                 "count": count_new_controller, "total_cost_usd": total_cost_new_controller}
    )
    
    # if there are setting changes
    count_setting_changes = cap_upgrades_df[capcontrol_upgrade_fields["change_cap_control"]].sum()
    unit_cost_setting_changes = controls_cost_database.loc[controls_cost_database["type"] ==
                                                           cost_database_fields["change_cap_control"]]["cost"].values[0]
    total_cost_setting_changes = count_setting_changes * unit_cost_setting_changes
    cap_cost.append( {"type": CapacitorControllerResultType.change_cap_control.value,
                     "count": count_setting_changes, "total_cost_usd": total_cost_setting_changes}
    )
    cap_cost_df = pd.DataFrame(cap_cost)
    cap_cost_df["comment"] = ""
    cap_cost_df = cap_cost_df[output_columns_list]
    return cap_cost_df


def compute_voltage_regcontrol_cost(voltage_upgrades_df, vreg_control_cost_database, vreg_xfmr_cost_database, xfmr_cost_database, keyword="RegControl"):
    """This function computes the voltage regulator controller related costs.
    Considered here: new voltage regulator controllers, control setting changes

    Parameters
    ----------
    voltage_upgrades_df
    controls_cost_database

    Returns
    -------

    """
    output_cost_field = "total_cost_usd"
    output_count_field = "count"
    output_columns_list = ["type", output_count_field, output_cost_field, "comment"]
    upgrade_fields_dict = {"add_new_reg_control": "new_controller_added",
                      "change_reg_control": "controller_settings_modified",
                      "add_new_transformer": "new_transformer_added",
                      "at_substation": "at_substation",
                    #   "change_ltc_control": "Substation LTC settings modified",
                        }
    cost_database_fields_dict = {"add_new_reg_control": "Add new voltage regulator controller",
                            "change_reg_control": "Change voltage regulator controller settings",
                            # "replace_reg_control": "Replace voltage regulator controller",   # this is not used currently
                            "add_substation_ltc": "Add new LTC controller",
                            "change_ltc_control": "Change LTC settings",
                            "add_new_transformer": "Add new voltage regulator transformer"}
    
    type_fields_dict = {i.name: i.value for i in VoltageRegulatorResultType}
    type_rows = list(type_fields_dict.keys())
    
    control_computation_fields = ["add_new_reg_control", "change_reg_control"]
    xfmr_fields = ["add_new_transformer"]

    empty_reg_cost_dict = {"type": type_rows, "count": [0] * len(type_rows),  "total_cost_usd": [0] * len(type_rows)}
    zero_cost_df = pd.DataFrame.from_dict(empty_reg_cost_dict)

    if voltage_upgrades_df.empty:  # if there are no voltage upgrades
        return zero_cost_df
    reg_upgrades_df = voltage_upgrades_df.loc[voltage_upgrades_df['equipment_type'].str.contains(keyword)]
    if reg_upgrades_df.empty:  # if there are no regulator controller upgrades
        return zero_cost_df

    cost_list = []
    # if there are regulator controller upgrades
    for field in control_computation_fields:
        # if at substation
        at_substation_df = reg_upgrades_df.loc[reg_upgrades_df[upgrade_fields_dict["at_substation"]] == True]
        if field == "add_new_reg_control":
            cost_field = "add_substation_ltc"
        elif field == "change_reg_control":
            cost_field = "change_ltc_control"
        else:
            raise Exception(f"Unknown field {field} in regulator cost computation")
        if not at_substation_df.empty:  # if there are no regcontrols at substation
            count = at_substation_df[upgrade_fields_dict[field]].sum()
            unit_cost = vreg_control_cost_database.loc[vreg_control_cost_database["type"] == cost_database_fields_dict[cost_field]]["cost"].values[0]
            total_cost = count * unit_cost
            cost_list.append({"type": type_fields_dict[cost_field], "count": count, "total_cost_usd": total_cost, "comment": ""})
        
        # if not at substation
        not_at_substation_df = reg_upgrades_df.loc[reg_upgrades_df[upgrade_fields_dict["at_substation"]] == False]
        cost_field = field
        if not not_at_substation_df.empty:  # if not at substation
            count = not_at_substation_df[upgrade_fields_dict[field]].sum()
            unit_cost = vreg_control_cost_database.loc[vreg_control_cost_database["type"] == cost_database_fields_dict[cost_field]]["cost"].values[0]
            total_cost = count * unit_cost
            cost_list.append({"type": type_fields_dict[cost_field], "count": count, "total_cost_usd": total_cost, "comment": ""})
    
    # add costs for added transformers (needed for voltage regulators)
    vreg_xfmr_cost_database = vreg_xfmr_cost_database.drop(columns=["type"])
    vreg_xfmr_cost_database = pd.concat([vreg_xfmr_cost_database, xfmr_cost_database])
    for field in xfmr_fields:
        cost_field = field
        # if at substation
        new_xfmr_added_df = reg_upgrades_df.loc[reg_upgrades_df[upgrade_fields_dict["add_new_transformer"]] == True]    
        for index, row in new_xfmr_added_df.iterrows():
            output_row = {}
            added_xfmr_details = row["final_settings"]
            added_xfmr_details = reformat_xfmr_upgrades_file(pd.DataFrame([added_xfmr_details]))
            deciding_columns = ["rated_kVA", "phases", "primary_kV", "secondary_kV", "primary_connection_type",
                        "secondary_connection_type", "num_windings"]
            params_dict = added_xfmr_details[["name"] + deciding_columns].to_dict(orient="records")[0]
            added_xfmr_details = added_xfmr_details.to_dict(orient="records")[0]
            output_row["equipment_parameters"] = params_dict
            # reformat xfmr dict
            unit_cost = vreg_xfmr_cost_database.loc[(vreg_xfmr_cost_database["rated_kVA"] == added_xfmr_details["rated_kVA"]) &
                                            (vreg_xfmr_cost_database["primary_kV"] == added_xfmr_details["primary_kV"]) &
                                            (vreg_xfmr_cost_database["secondary_kV"] == added_xfmr_details["secondary_kV"]) &
                                            (vreg_xfmr_cost_database["phases"] == added_xfmr_details["phases"]) &
                                            (vreg_xfmr_cost_database["num_windings"] == added_xfmr_details["num_windings"]) &
                                            (vreg_xfmr_cost_database["primary_connection_type"] == added_xfmr_details["primary_connection_type"]) &
                                            (vreg_xfmr_cost_database["secondary_connection_type"] == added_xfmr_details["secondary_connection_type"])
                                            ]["cost"]
            if len(unit_cost) > 0:  # if there are more than one rows, the first one is chosen in random
                unit_cost = unit_cost.values[0]
                output_row[output_cost_field] = unit_cost
                output_row["comment"] = ""
                output_row[output_count_field] = 1
            else:  # if costs are not present for this transformer, then choose from other xfmr database rated_kVA
                backup_deciding_property = "rated_kVA"
                closest = vreg_xfmr_cost_database.loc[abs(vreg_xfmr_cost_database[backup_deciding_property] - added_xfmr_details[backup_deciding_property]).idxmin()].copy()
                if isinstance(closest, pd.DataFrame):
                    closest = closest.iloc[0]
                output_row[output_cost_field] = closest["cost"]
                output_row[output_count_field] = 1
                for key, value in closest.items():  # make it json serializable
                    if isinstance(value, np.int64):
                        closest[key] = int(value)
                comment_string =  {"text": f"Transformer {added_xfmr_details['name']}: Exact cost not available. " \
                                           f"Unit cost for transformer with these parameters used " \
                                           f"(based on closest {backup_deciding_property}", 
                                  "params": dict(closest)}
                                  
                logger.debug(comment_string)
                output_row["comment"] = comment_string
            if row[upgrade_fields_dict["at_substation"]]:
                output_row["type"] = type_fields_dict["add_new_substation_transformer"]
            else: 
                output_row["type"] = type_fields_dict["add_new_vreg_transformer"]
            cost_list.append(output_row)

    reg_cost_df = pd.DataFrame(cost_list)
    reg_cost_df = reg_cost_df[output_columns_list]
    return reg_cost_df


def get_total_costs(thermal_cost_df, voltage_cost_df):
    """This function combines voltage and thermal upgrades costs into one file.
    """
    total_cost_df = pd.concat([thermal_cost_df, voltage_cost_df])
    total_cost_df = total_cost_df.groupby('type').sum(numeric_only=True)
    total_cost_df.reset_index(inplace=True)
    return total_cost_df


if __name__ == "__main__":
    compute_all_costs()




from disco.distribution.deployment_parameters import DeploymentParameters


class UpgradeParameters(DeploymentParameters):
    """Represents deployment parameters for upgrade simulation/analysis

    Parameters
    ----------
    DeploymentParameters : object
        Deployment Job
    """





import click

from disco.cli.config_snapshot import snapshot
from disco.cli.config_time_series import time_series
from disco.cli.config_upgrade import upgrade


@click.group()
def config():
    """Create JADE configurations for DISCO analysis types"""


config.add_command(snapshot)
config.add_command(time_series)
config.add_command(upgrade)



import logging

import click

from jade.loggers import setup_logging
from disco.pipelines.utils import ensure_jade_pipeline_output_dir
from disco.postprocess.plots import plot_voltage, plot_hc

logger = logging.getLogger(__name__)


@click.command()
@click.argument("output_dir")
@click.option(
    "--scenario",
    type=click.STRING,
    default="scenario",
    show_default=True,
    help="Input the PyDSS scenario"
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def plot(output_dir, scenario, verbose):
    """Plot voltage and hosting capacity charts"""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])
    output_dir = ensure_jade_pipeline_output_dir(output_dir)
    plot_voltage(output_dir, scenario)
    plot_hc(output_dir, scenario)
